-----------------------------------------------------------------
----- LOCALIZED FUNCTIONS ---------------------------------------
-----------------------------------------------------------------
local sins, coss, cur_obj_resolve_wall_collisions, obj_mark_for_deletion, network_player_from_global_index,
 obj_scale, calculate_pitch, obj_move_xyz, spawn_non_sync_object = 
sins, coss, cur_obj_resolve_wall_collisions, obj_mark_for_deletion, network_player_from_global_index,
 obj_scale, calculate_pitch, obj_move_xyz, spawn_non_sync_object

-----------------------------------------------------------
--- EXTRA FIELDS ------------------------------------------
-----------------------------------------------------------

define_custom_obj_fields({
    oOwner = 'u32',
    oShotLevel = 'f32',
    oShotChar = 'u32',  -- 0 is X, 1 is Vile
    oLightArmorShotOffset = 'u32',
    oShotDestroyNextFrame = 'u32',
    oChargeSize = 'f32',
    -- Also used for light armor shot initial phase
    oChargeRot = 'u32',
    oChargeMinSize = 'f32',
    oChargeScaleDownFactor = 'f32',
    oChargeIndex = 'u32',
})

-----------------------------------------------------------
--- FUNCTIONS ---------------------------------------------
-----------------------------------------------------------

function spawn_charge(m, charge_level)
    local mGfx = m.marioObj.header.gfx
    if (charge_level == 1) then
        spawn_non_sync_object(id_bhvChargeSpheres, E_model_energy_charge1, mGfx.pos.x, mGfx.pos.y + 80, mGfx.pos.z, function (obj)
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oMoveAngleRoll = 0
            obj.oChargeIndex = 1
        end)
    elseif (charge_level == 2) then
        spawn_non_sync_object(id_bhvCharge2Back, E_model_charge2_back, mGfx.pos.x, mGfx.pos.y + 80, mGfx.pos.z, function (obj)
            obj.oChargeMinSize = 1.1
            obj.oChargeScaleDownFactor = 0.14
            obj.oChargeSize = 2.1
            obj.oChargeRot = 300
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oChargeIndex = 2
        end)
        spawn_non_sync_object(id_bhvCharge2Back, E_model_charge2_front, mGfx.pos.x, mGfx.pos.y + 80, mGfx.pos.z, function (obj)
            obj.oChargeMinSize = 1.6
            obj.oChargeScaleDownFactor = 0.09
            obj.oChargeSize = 2.8
            obj.oChargeRot = 100
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oChargeIndex = 2
        end)
        spawn_non_sync_object(id_bhvChargeSpheres, E_model_energy_charge2, mGfx.pos.x, mGfx.pos.y + 80, mGfx.pos.z, function (obj)
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oMoveAngleRoll = 0
            obj.oChargeIndex = 2
        end)
    elseif (charge_level == 3) then
        spawn_non_sync_object(id_bhvCharge2Back, E_model_charge3_back_lightarmor, mGfx.pos.x, mGfx.pos.y + 80, mGfx.pos.z, function (obj)
            obj.oChargeMinSize = 1.1
            obj.oChargeScaleDownFactor = 0.14
            obj.oChargeSize = 2.1
            obj.oChargeRot = 300
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oChargeIndex = 3
        end)
        spawn_non_sync_object(id_bhvCharge2Back, E_model_charge3_front_lightarmor, mGfx.pos.x, mGfx.pos.y + 80, mGfx.pos.z, function (obj)
            obj.oChargeMinSize = 1.6
            obj.oChargeScaleDownFactor = 0.09
            obj.oChargeSize = 2.8
            obj.oChargeRot = 100
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oChargeIndex = 3
        end)
        spawn_non_sync_object(id_bhvChargeSpheres, E_model_energy_charge3_lightarmor, mGfx.pos.x, mGfx.pos.y + 80, mGfx.pos.z, function (obj)
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oMoveAngleRoll = 0
            obj.oChargeIndex = 3
        end)
    end
end

-----------------------------------------------------------
--- HELPERS -----------------------------------------------
-----------------------------------------------------------

-- Replacement for native create_sound_spawner()
function create_sound_spawner(obj, soundId)
    local soundObj = spawn_non_sync_object(id_bhvSoundSpawner, 0, 0, 0, 0, nil);
    if soundObj == nil then
        return
    end
    obj_copy_pos_and_angle(soundObj, obj)
    soundObj.oSoundEffectUnkF4 = soundId
end

-----------------------------------------------------------
--- PARTICLES ---------------------------------------------
-----------------------------------------------------------

--- DASH TRAILS
local trail_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE)
    o.oOpacity = 90
    cur_obj_hide()
end

local trail_loop = function(o)
    if cur_obj_resolve_wall_collisions() == 0 then
        obj_move_xyz(o, o.oVelX, o.oVelY, o.oVelZ)
    end

    if o.oTimer == 1 then
        cur_obj_unhide()
    end
    if o.oTimer >= 3 then
        o.oOpacity = o.oOpacity - 15
    end

    if o.oOpacity <= 0 then
        obj_mark_for_deletion(o)
    end
end



--- CHARGE 2 EFFECTS (highest charge level)
-- front
local charge2_front_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE)
    o.oOpacity = 0xA9
    obj_scale(o, 1.75)
end

local charge2_front_loop = function(o)
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    local s = gPlayerSyncTable[m.playerIndex]
    local mGfx = m.marioObj.header.gfx
    o.oPosX = mGfx.pos.x
    o.oPosY = mGfx.pos.y + 90 - MODEL_Y_OFFSET
    o.oPosZ = mGfx.pos.z
    o.oMoveAngleYaw = m.faceAngle.y - 16384 + 4096 * o.oTimer
    if o.oTimer >= 8 then
        obj_mark_for_deletion(o)
    end  
end

-- back
local charge2_back_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW)
    o.oOpacity = 255
    o.header.gfx.scale.x = o.oChargeSize
    o.header.gfx.scale.y = o.oChargeSize
    o.header.gfx.scale.z = o.oChargeSize
    cur_obj_hide()
end

local charge2_back_loop = function(o)
    if (o.oTimer == 1) then
        cur_obj_unhide()
    end
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    local s = gPlayerSyncTable[m.playerIndex]
    local mGfx = m.marioObj.header.gfx

    if (o.header.gfx.scale.x >= o.oChargeMinSize) then
        o.header.gfx.scale.x = o.header.gfx.scale.x - o.oChargeScaleDownFactor
        o.header.gfx.scale.y = o.header.gfx.scale.y - o.oChargeScaleDownFactor
        o.header.gfx.scale.z = o.header.gfx.scale.z - o.oChargeScaleDownFactor
    else
        o.header.gfx.scale.x = o.oChargeSize
        o.header.gfx.scale.y = o.oChargeSize
        o.header.gfx.scale.z = o.oChargeSize
    end

    local m0 = gMarioStates[0]
    o.oMoveAngleYaw = m0.area.camera.yaw

    local pitch = calculate_pitch(m0.pos, m0.area.camera.pos)
    o.oMoveAnglePitch = -pitch
    
    o.oMoveAngleRoll = o.oMoveAngleRoll + o.oChargeRot

    o.oPosX = mGfx.pos.x
    o.oPosY = mGfx.pos.y + 90 - MODEL_Y_OFFSET
    o.oPosZ = mGfx.pos.z
    
    if (s.x_charge_spawned ~= o.oChargeIndex) then
        obj_mark_for_deletion(o)
    end  
end

--- CHARGE SPHERES (both charge levels)
local charge_spheres_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW)
    --o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_BILLBOARD
    o.header.gfx.scale.x = 2.4
    o.header.gfx.scale.y = 2.4
    o.header.gfx.scale.z = 2.4
    if (o.oMoveAngleRoll == 8192) then
        o.header.gfx.scale.x = 2.15
        o.header.gfx.scale.y = 2.15
        o.header.gfx.scale.z = 2.15
    end
    cur_obj_hide()
end

local charge_spheres_loop = function(o)
    if (o.oTimer == 1) then
        cur_obj_unhide()
    end
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    local m0 = gMarioStates[0]
    local s = gPlayerSyncTable[m.playerIndex]

    -- CHECK IN MULTIPLAYER
    o.oMoveAngleYaw = m0.area.camera.yaw
    local pitch = calculate_pitch(m0.pos, m0.area.camera.pos)
    o.oMoveAnglePitch = -pitch
    
    --o.oMoveAngleRoll = o.oMoveAngleRoll + 700
    local mGfx = m.marioObj.header.gfx
    o.oPosX = mGfx.pos.x
    o.oPosY = mGfx.pos.y + 95 - MODEL_Y_OFFSET
    o.oPosZ = mGfx.pos.z
    o.header.gfx.scale.x = o.header.gfx.scale.x - 0.175
    o.header.gfx.scale.y = o.header.gfx.scale.y - 0.175
    o.header.gfx.scale.z = o.header.gfx.scale.z - 0.175
    if o.oTimer >= 10 then
        cur_obj_hide()
    end

    if (s.x_charge_spawned ~= o.oChargeIndex) or o.oTimer >= 11 then
        obj_mark_for_deletion(o)
    end 

    if o.oTimer == 6 then

        if (o.oChargeIndex == 1) then
            spawn_non_sync_object(id_bhvChargeSpheres, E_model_energy_charge1, mGfx.pos.x, mGfx.pos.y + 90 - MODEL_Y_OFFSET, mGfx.pos.z, function (obj)
                obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
                if o.oMoveAngleRoll == 8192 then
                    obj.oMoveAngleRoll = 0
                else
                    obj.oMoveAngleRoll = 8192
                end
                obj.oChargeIndex = 1
            end)
        elseif (o.oChargeIndex == 2) then
            spawn_non_sync_object(id_bhvChargeSpheres, E_model_energy_charge2, mGfx.pos.x, mGfx.pos.y + 90 - MODEL_Y_OFFSET, mGfx.pos.z, function (obj)
                obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
                if o.oMoveAngleRoll == 8192 then
                    obj.oMoveAngleRoll = 0
                else
                    obj.oMoveAngleRoll = 8192
                end
                obj.oChargeIndex = 2
            end)
        elseif (o.oChargeIndex == 3) then
            spawn_non_sync_object(id_bhvChargeSpheres, E_model_energy_charge3_lightarmor, mGfx.pos.x, mGfx.pos.y + 90 - MODEL_Y_OFFSET, mGfx.pos.z, function (obj)
                obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
                if o.oMoveAngleRoll == 8192 then
                    obj.oMoveAngleRoll = 0
                else
                    obj.oMoveAngleRoll = 8192
                end
                obj.oChargeIndex = 3
            end)
        end
    end
end



--- Legacy charge1 code
local charge1_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE)
    o.header.gfx.scale.x = 2
    o.header.gfx.scale.y = 2;
    o.header.gfx.scale.z = 2;
end

local charge1_loop = function(o)
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    local s = gPlayerSyncTable[m.playerIndex]
    local mGfx = m.marioObj.header.gfx
    o.oPosX = mGfx.pos.x
    o.oPosY = mGfx.pos.y + 90 - MODEL_Y_OFFSET
    o.oPosZ = mGfx.pos.z
    if o.oTimer >= 1 then
        cur_obj_hide()
    end
    if o.oTimer == 2 then
        spawn_non_sync_object(id_bhvCharge2, E_model_charge2, mGfx.pos.x, mGfx.pos.y + 90 - MODEL_Y_OFFSET, mGfx.pos.z, function (obj)
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oMoveAngleYaw = m.faceAngle.y
        end)
    end
    if o.oTimer >= 10 or s.charge_timer == 0 then
        obj_mark_for_deletion(o)
    end 
end



--- WALL SPARK
local wallspark_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE)
end

local wallspark_loop = function(o)
    obj_scale(o, o.header.gfx.scale.x + 0.55)

    if o.oTimer > 4 then
        obj_mark_for_deletion(o)
    end
end



--- SHOT RING
local shot_ring_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE)
    o.oOpacity = 255
    obj_scale(o, 0.3)
end

local shot_ring_loop = function(o)

    o.oVelX = o.oForwardVel * sins(o.oMoveAngleYaw) * coss(o.oMoveAnglePitch)
    o.oVelY = o.oForwardVel * sins(o.oMoveAnglePitch)
    o.oVelZ = o.oForwardVel * coss(o.oMoveAngleYaw) * coss(o.oMoveAnglePitch)

    if cur_obj_resolve_wall_collisions() == 0 then
        obj_move_xyz(o, o.oVelX, o.oVelY, o.oVelZ)
    end

    obj_scale(o, o.header.gfx.scale.x + 0.45)

    if o.oTimer >= 3 then
        o.oOpacity = o.oOpacity - 12
    end

    if o.oTimer > 4 then
        obj_mark_for_deletion(o)
    end
end



--- SHOT TRAIL
local shot_trail_ball_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE)
    o.oOpacity = 255
    if (o.oShotLevel == 4) then
        obj_scale(o, 1.1)
    else
        obj_scale(o, 0.7)
    end
end

local shot_trail_ball_loop = function(o)

    if cur_obj_resolve_wall_collisions() == 0 then
        obj_move_xyz(o, o.oVelX, o.oVelY, o.oVelZ)
    end

    if (o.oShotLevel == 4) then
        obj_scale(o, o.header.gfx.scale.x - 0.08)
        if o.oTimer > 5 then
            obj_mark_for_deletion(o)
        end
    else
        obj_scale(o, o.header.gfx.scale.x - 0.1)
        if o.oTimer > 4 then
            obj_mark_for_deletion(o)
        end
    end
end




-------------------------------------------------------
--- WEAPONS -------------------------------------------
-------------------------------------------------------






--- CHERRY BLAST

--- SHOULDER CANNON


--------------------------------------------------
--- REGISTRING BEHAVIORS -------------------------
--------------------------------------------------


id_bhvTrail = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, trail_init, trail_loop)
id_bhvCharge2Front = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, charge2_front_init, charge2_front_loop)
id_bhvCharge2Back = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, charge2_back_init, charge2_back_loop)
id_bhvChargeSpheres = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, charge_spheres_init, charge_spheres_loop)
id_bhvCharge1 = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, charge1_init, charge1_loop)
id_bhvShotRing = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, shot_ring_init, shot_ring_loop)
id_bhvShotTrailBall = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, shot_trail_ball_init, shot_trail_ball_loop)
id_bhvWallspark = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, wallspark_init, wallspark_loop)
