-----------------------------------------------------------------
----- LOCALIZED FUNCTIONS ---------------------------------------
-----------------------------------------------------------------
local math_sin, math_cos, mario_check_object_grab, mario_floor_is_steep, mario_drop_held_object, spawn_non_sync_object, analog_stick_held_back, apply_slope_accel,
 set_mario_action, lava_boost_on_wall, check_fall_damage_or_get_stuck, perform_air_step, smlua_anim_util_set_animation, approach_s32, should_begin_sliding,
  update_air_without_turn, common_air_action_step, mario_set_forward_vel, begin_braking_action, set_jump_from_landing, smlua_anim_util_set_animation,
  set_mario_animation, align_with_floor, mmx_play_sound =
 math.sin, math.cos, mario_check_object_grab, mario_floor_is_steep, mario_drop_held_object, spawn_non_sync_object, analog_stick_held_back, apply_slope_accel,
  set_mario_action, lava_boost_on_wall, check_fall_damage_or_get_stuck, perform_air_step, smlua_anim_util_set_animation, approach_s32, should_begin_sliding,
   update_air_without_turn, common_air_action_step, mario_set_forward_vel, begin_braking_action, set_jump_from_landing, smlua_anim_util_set_animation,
   set_mario_animation, align_with_floor, mmx_play_sound

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
---- ACTIONS -----------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------


-------------------------------------------
--- ALLOCATIONS ---------------------------
-------------------------------------------

ACT_AIR_HIT_WALL_CUSTOM =       allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_MOVING | ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION | ACT_FLAG_CONTROL_JUMP_HEIGHT)
ACT_DASH =                      allocate_mario_action(ACT_GROUP_MOVING | ACT_FLAG_MOVING | ACT_FLAG_BUTT_OR_STOMACH_SLIDE)
ACT_AIR_DASH =             allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_MOVING | ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION)
ACT_JUMP_CUSTOM =           allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_MOVING | ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION | ACT_FLAG_CONTROL_JUMP_HEIGHT)
ACT_FREEFALL_CUSTOM =           allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_MOVING | ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION)
ACT_WALL_SLIDE =           allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_MOVING | ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION)
ACT_WALL_KICK_CUSTOM =      allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_MOVING | ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION | ACT_FLAG_CONTROL_JUMP_HEIGHT)
ACT_WALKING_CUSTOM =               allocate_mario_action(ACT_GROUP_MOVING | ACT_FLAG_MOVING | ACT_FLAG_ALLOW_FIRST_PERSON)


-------------------------------------------
--- RESOURCES -----------------------------
-------------------------------------------

local SOUND_X_DASH = SOUND_X_DASH
local SOUND_X_WALLHIT = SOUND_X_WALLHIT

-------------------------------------------
--- STEP FUNCTIONS ------------------------
-------------------------------------------

-- This is like common_air_action_step except that there is no wall interaction
-- Usually called when the vspeed is positive, to prevent the wall slide action with positive vspeed
function common_air_action_step_no_wall(m, landAction, animation, stepArg)
    if not m then return 0 end
    local stepResult

    update_air_without_turn(m)

    stepResult = perform_air_step(m, stepArg)
    if m.action == ACT_BUBBLED and stepResult == AIR_STEP_HIT_LAVA_WALL then
        stepResult = AIR_STEP_HIT_WALL
    end

    if stepResult == AIR_STEP_NONE then
        set_character_animation(m, animation)
    elseif stepResult == AIR_STEP_LANDED then
        if check_fall_damage_or_get_stuck(m, ACT_HARD_BACKWARD_GROUND_KB) == 0 then
            set_mario_action(m, landAction, 0)
        end
    elseif stepResult == AIR_STEP_GRABBED_CEILING then
        set_mario_action(m, ACT_START_HANGING, 0)
    elseif stepResult == AIR_STEP_HIT_LAVA_WALL then
        lava_boost_on_wall(m)
    end

    return stepResult
end

local function limit_angle(a)
    return (a + 0x8000) % 0x10000 - 0x8000
end

local function update_walking_speed_mmx(m)
    local s = gPlayerSyncTable[m.playerIndex]
    local maxTargetSpeed = 0
    local targetSpeed    = 0
    local char_accel = 1.125

    if m.floor ~= 0 and m.floor.type == SURFACE_SLOW then
        maxTargetSpeed = 24.0
    else
        if (s.character == 'vile') then
            maxTargetSpeed = 43.0
        else
            maxTargetSpeed = 32.0
        end
    end

    local intendedMag = m.intendedMag
    if (s.character == 'vile') then
        intendedMag = intendedMag * 1.34375
        char_accel = 1.66
    end

    if intendedMag < maxTargetSpeed then
        targetSpeed = intendedMag
    else
        targetSpeed = maxTargetSpeed
    end

    if m.quicksandDepth > 10.0 then
        targetSpeed = targetSpeed * (6.25 / m.quicksandDepth)
    end

    if m.forwardVel <= 0.0 then
        m.forwardVel = m.forwardVel + char_accel
    elseif m.forwardVel <= targetSpeed then
        m.forwardVel = m.forwardVel + (char_accel - m.forwardVel / 43.0)
    elseif m.floor ~= 0 and m.floor.normal.y >= 0.95 then
        m.forwardVel = m.forwardVel - 1.0
    end

    if m.forwardVel > 48.0 then
        m.forwardVel = 48.0
    end

    m.faceAngle.y = m.intendedYaw
    
    apply_slope_accel(m)
end

-- unused for now
local function anim_and_audio_for_walk_mmx(m)
    if not m then return end
    local val14 = 0
    local marioObj = m.marioObj
    local val0C = true
    local targetPitch = 0
    local val04 = 0

    val04 = m.intendedMag > m.forwardVel and m.intendedMag or m.forwardVel

    if val04 < 4.0 then
        val04 = 4.0
    end

    if m.quicksandDepth > 50.0 then
        val14 = math.floor(val04 / 4.0 * 0x10000)
        set_character_anim_with_accel(m, CHAR_ANIM_MOVE_IN_QUICKSAND, val14)
        play_step_sound(m, 19, 93)
        m.actionTimer = 0
    else
        while val0C do
            if m.actionTimer == 0 then
                if val04 > 8.0 then
                    m.actionTimer = 2
                else
                    if val04 / 4.0 * 0x10000 < 0x1000 then
                        val14 = 0x1000
                    else
                        val14 = math.floor(val04 / 4.0 * 0x10000)
                    end
                    set_character_anim_with_accel(m, CHAR_ANIM_START_TIPTOE, val14)
                    play_step_sound(m, 7, 22)
                    if is_anim_past_frame(m, 23) then
                        m.actionTimer = 2
                    end
                    val0C = false
                end
            elseif m.actionTimer == 1 then
                if val04 > 8.0 then
                    m.actionTimer = 2
                else
                    val14 = math.floor(val04 * 0x10000)
                    if val14 < 0x1000 then
                        val14 = 0x1000
                    end
                    set_character_anim_with_accel(m, CHAR_ANIM_TIPTOE, val14)
                    play_step_sound(m, 14, 72)
                    val0C = false
                end
            elseif m.actionTimer == 2 then
                if val04 < 5.0 then
                    m.actionTimer = 1
                elseif val04 > 22.0 then
                    m.actionTimer = 3
                else
                    val14 = math.floor(val04 / 4.0 * 0x10000)
                    set_character_anim_with_accel(m, CHAR_ANIM_WALKING, val14)
                    play_step_sound(m, 10, 49)
                    val0C = false
                end
            elseif m.actionTimer == 3 then
                if val04 < 18.0 then
                    m.actionTimer = 2
                else
                    val14 = math.floor(val04 / 4.0 * 0x10000)
                    set_character_anim_with_accel(m, CHAR_ANIM_RUNNING, val14)
                    --play_step_sound(m, 9, 45)
                    if (m.marioObj.header.gfx.animInfo.animFrame == 9 or m.marioObj.header.gfx.animInfo.animFrame == 10) then
                        --mmx_play_sound(SOUND_VILE_STEP, m.pos, 0.57)
                    end
                    if (is_anim_past_frame(m, 9) ~= 0 and is_anim_past_frame(m, 13) == 0) or 
                    (is_anim_past_frame(m, 45) ~= 0 and is_anim_past_frame(m, 49) == 0) then
                        mmx_play_sound(SOUND_VILE_STEP, m.pos, 0.3)
                    end
                    targetPitch = tilt_body_running(m)
                    val0C = false
                end
            else
                val0C = false
            end
        end
    end

    marioObj.oMarioWalkingPitch = approach_s32(marioObj.oMarioWalkingPitch, targetPitch, 0x800, 0x800)
    marioObj.header.gfx.angle.x = marioObj.oMarioWalkingPitch
end


----------------------------------------------
--- DEFINITION -------------------------------
----------------------------------------------

--- WALKING CUSTOM
local function act_walking_custom(m)
    local startPos = m.pos
    local startYaw = m.faceAngle.y

    mario_drop_held_object(m)

    if should_begin_sliding(m) ~= 0 then
        return set_mario_action(m, ACT_BEGIN_SLIDING, 0)
    end

    if (m.input & INPUT_FIRST_PERSON) ~= 0 then
        return begin_braking_action(m)
    end

    if (m.input & INPUT_A_PRESSED) ~= 0 then
        return set_jump_from_landing(m)
    end

    if (m.input & INPUT_ZERO_MOVEMENT) ~= 0 then
        return begin_braking_action(m)
    end

    

    --if (m.input & INPUT_Z_PRESSED) ~= 0 then
    --    return set_mario_action(m, ACT_CROUCH_SLIDE, 0)
    --end

    m.actionState = 0

    vec3f_copy(startPos, m.pos)
    update_walking_speed_mmx(m)

    local stepResult = perform_ground_step(m)
    
    if stepResult == GROUND_STEP_LEFT_GROUND then
        set_mario_action(m, ACT_FREEFALL, 0)
        set_mario_animation(m, MARIO_ANIM_GENERAL_FALL)
    elseif stepResult == GROUND_STEP_NONE then
        anim_and_audio_for_walk(m)
        if (m.intendedMag - m.forwardVel > 16.0) then
            set_mario_particle_flags(m, PARTICLE_DUST, false);
        end
    elseif stepResult == GROUND_STEP_HIT_WALL then
        push_or_sidle_wall(m, startPos)
        m.actionTimer = 0
    end
    --tilt_body_walking(m, startYaw)
    return false
end

--- JUMP CUSTOM
-- This is basically the same act_jump code found in those c files but without the ledge grab check
-- Don't know if there was an alternative way to do this
function act_jump_custom(m)
    local s = gPlayerSyncTable[m.playerIndex]
    if (not m) then
        return 0
    end
    
    -- Not needed

    --if (check_kick_or_dive_in_air(m) ~= 0) then
    --    return true
    --end

    if (m.input & INPUT_A_PRESSED) ~= 0 and m.actionTimer > 0 and (m.flags & MARIO_WING_CAP) ~= 0 then
        if (m.playerIndex == 0) then
            s.already_airdashed = false
            s.jump_dash_state = false
        end
        m.vel.y = 38
        return set_mario_action(m, ACT_FLYING_TRIPLE_JUMP, 0)
    end

    -- Variable Air Dash not available (v1.3)

    --if (m.input & INPUT_A_PRESSED) ~= 0 and m.actionTimer > 0 then
    --    return set_mario_action(m, ACT_VAD_UP, 0)
    --end

    play_mario_sound(m, SOUND_ACTION_TERRAIN_JUMP, 0)

    -- ACT_JUMP_LAND is trash?

    if (m.vel.y >= 0) then
        common_air_action_step_no_wall(m, ACT_FREEFALL_LAND_STOP, MARIO_ANIM_SINGLE_JUMP, AIR_STEP_CHECK_HANG)
    else
        common_air_action_step(m, ACT_FREEFALL_LAND_STOP, MARIO_ANIM_SINGLE_JUMP, AIR_STEP_CHECK_HANG)
    end
    m.actionTimer = m.actionTimer + 1
    return false
end



--- FREEFALL_CUSTOM
function act_freefall_custom(m)
    if (not m) then
        return 0
    end

    if (m.vel.y >= 0) then
        common_air_action_step_no_wall(m, ACT_FREEFALL_LAND_STOP, MARIO_ANIM_SINGLE_JUMP, AIR_STEP_CHECK_HANG)
    else
        common_air_action_step(m, ACT_FREEFALL_LAND_STOP, MARIO_ANIM_SINGLE_JUMP, AIR_STEP_CHECK_HANG)
    end
    return false
end




--- WALL KICK CUSTOM
function act_wall_kick_custom(m)
    local s = gPlayerSyncTable[m.playerIndex]
    local maxVelocity = 32
    local kickvspeed = 44.75
    if (m.actionTimer == 0) then
        if (s.character == 'x') then
            mmx_play_sound(SOUND_X_WALLKICK, m.pos, 0.3)
        elseif (s.character == 'vile') then
            mmx_play_sound(SOUND_VILE_JUMP, m.pos, 0.4)
        end
    end
    if (s.character == 'vile') then
        maxVelocity = 43
        kickvspeed = 45.3
    end
    local mGfx = m.marioObj.header.gfx
    if s.jump_dash_state then
        maxVelocity = 54
    end
    if s.firstWallKick then
        if (m.playerIndex == 0) then
            s.firstWallKick = false
            s.quasiSameSideWallKicks = 0
        end
    end
    if (m.actionTimer == 0) then
        local vspeed
        if s.jump_dash_state then
            vspeed = 46.25
        else
            vspeed = kickvspeed
        end
        if s.quasiSameSideWallKicks < 1 then
            spawn_non_sync_object(id_bhvWallspark, E_model_wallspark, mGfx.pos.x , mGfx.pos.y , mGfx.pos.z , function (obj)
                obj.oMoveAngleYaw = m.faceAngle.y + 32768
            end)
        end
        vspeed = vspeed - s.quasiSameSideWallKicks * 2.5
        if (vspeed <= 0) then
            vspeed = 0
        end
        m.vel.y = vspeed
        m.faceAngle.y = m.faceAngle.y + 32768
        if (m.playerIndex == 0) then
            s.initialFaceAngleAfterWallKick = m.faceAngle.y
        end
        mario_set_forward_vel(m, -maxVelocity)
    end
    if (m.actionTimer <= 6) then
        local initialFaceTheta = (s.initialFaceAngleAfterWallKick / 0x8000) * math.pi
        m.vel.x = math_sin(initialFaceTheta)
        m.vel.z = math_cos(initialFaceTheta)
    else
        local normalStickMag = m.controller.stickMag / 64
        mario_set_forward_vel(m, maxVelocity * normalStickMag)
    end

    -- Air dashing in here is different because it's not just the current face angle
    if (s.character == 'x' and s.x_armor == 0) then
        if (m.input & INPUT_Z_PRESSED) ~= 0 and s.jump_dash_state == false then
            m.vel.y = 0
            -- If action timer is smaller than 6, the airdash direction would be based on the stick orientation
            -- otherwise we don't have to do this calculation
            if (m.actionTimer <= 6) then
                m.faceAngle.y = m.intendedYaw
            end
            return set_mario_action(m, ACT_AIR_DASH, 1)
        end
    end

    set_character_animation(m, CHAR_ANIM_SINGLE_JUMP)
    if s.x_shooting == 0 or s.vile_shooting == 0 then
        if m.actionTimer > 5 and s.character ~= 'none' then
            smlua_anim_util_set_animation(m.marioObj, s.character .. "_jump")
        else
            smlua_anim_util_set_animation(m.marioObj, "x_wall_kick")
        end
    end
    if m.actionTimer > 5 then
        if m.vel.y <= 0 then
            local velFrame = math.floor(m.vel.y * 0.18)
            if velFrame < -9 then
                velFrame = -9
            end
            local frameCounter = 12 - m.actionTimer
            if frameCounter < 0 then
                frameCounter = 0
            end
            set_anim_to_frame(m, 12 + velFrame + frameCounter)
        else
            set_anim_to_frame(m, 12)
        end
    else
        set_anim_to_frame(m, m.actionTimer)
    end
    m.actionTimer = m.actionTimer + 1

    if (m.vel.y > 2) then
        common_air_action_step_no_wall(m, ACT_FREEFALL_LAND_STOP, MARIO_ANIM_SINGLE_JUMP, AIR_STEP_CHECK_HANG)
    else
        common_air_action_step(m, ACT_FREEFALL_LAND_STOP, MARIO_ANIM_SINGLE_JUMP, AIR_STEP_CHECK_HANG)
    end
    return 0
end




--- AIR HIT WALL CUSTOM
function act_air_hit_wall_custom(m)
    local s = gPlayerSyncTable[m.playerIndex]
    local wallYaw = atan2s(m.wallNormal.z, m.wallNormal.x)
    if (m.playerIndex == 0) then
        s.jump_dash_state = false
    end
    if m.heldObj ~= 0 then
        mario_drop_held_object(m)
    end

    if m.wall then
        m.faceAngle.y = wallYaw
    end
    
    return set_mario_action(m, ACT_WALL_SLIDE, 0)
end




--- WALL SLIDE
function point_is_in_funnel(x1, z1, x2, z2, direction_angle, fov_angle)
    local angle = math.atan2(x1 - x2, z1 - z2)
    if angle < 0 then
        angle = angle + 2 * math.pi
    end
    angle_diff = math.abs(angle - direction_angle)
    
    if angle_diff > math.pi then
        angle_diff = 2 * math.pi - angle_diff
    end
    return angle_diff <= fov_angle / 2
end

function act_wall_slide(m)
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.playerIndex == 0 and s.already_airdashed) then
        s.already_airdashed = false
    end
    if (m.actionTimer == 0) then
        mmx_play_sound(SOUND_X_WALLHIT, m.pos, 0.5)
    end
    if s.x_shooting == 0 or s.vile_shooting == 0 then
        smlua_anim_util_set_animation(m.marioObj, "x_wall_slide")
    end
    --local e = gMarioStateExtras[m.playerIndex]
    --e.savedWallSlideHeight = m.pos.y
    --e.savedWallSlide = true
    -- m.faceAngle.y = limit_angle(m.faceAngle.y + 0x8000)
    if (m.input & INPUT_A_PRESSED) ~= 0 then

        local mmx = gMmxStates[m.playerIndex]
        local wallYaw = atan2s(m.wallNormal.z, m.wallNormal.x)

        local wallYawRadiants = (wallYaw / 0x8000) * math.pi
        if wallYawRadiants < 0 then
            wallYawRadiants = wallYawRadiants + 2 * math.pi
        end
        local wallYawDiff = math.abs(wallYawRadiants - mmx.previousWallYawRads)
        if wallYawDiff > math.pi then
            wallYawDiff = 2 * math.pi - wallYawDiff
        end
        local check_point = point_is_in_funnel(m.pos.x, m.pos.z, mmx.lastWallKickPos.x, mmx.lastWallKickPos.z, mmx.previousWallYawRads, math.pi / 2)
        if (m.playerIndex == 0) then
            if check_point and wallYawDiff >= math.pi / 1.9 then
                s.quasiSameSideWallKicks = 0
            else
                s.quasiSameSideWallKicks = s.quasiSameSideWallKicks + 1
            end
        end
        mmx.previousWallYawRads = wallYawRadiants
        mmx.lastWallKickPos = {x = m.pos.x, z = m.pos.z}

        if (m.playerIndex == 0) then
            s.just_walljumped = 6
            if (m.input & INPUT_Z_DOWN) ~= 0 then
                s.jump_dash_state = true
            end
        end
        -- m.faceAngle.y = limit_angle(m.faceAngle.y + 0x8000)
        return set_mario_action(m, ACT_WALL_KICK_CUSTOM, 1)
    end

    -- attempt to stick to the wall a bit. if it's 0, sometimes you'll get kicked off of slightly sloped walls
    mario_set_forward_vel(m, -1.0)

    m.particleFlags = m.particleFlags | PARTICLE_DUST

    play_sound(SOUND_MOVING_TERRAIN_SLIDE + m.terrainSoundAddend, m.marioObj.header.gfx.cameraToObject)
    
    if perform_air_step(m, 0) == AIR_STEP_LANDED then
        mario_set_forward_vel(m, 0.0)
        if check_fall_damage_or_get_stuck(m, ACT_HARD_BACKWARD_GROUND_KB) == 0 then
            return set_mario_action(m, ACT_JUMP_LAND, 0)
        end
    end

    if m.actionTimer < 6 then
        set_anim_to_frame(m, m.actionTimer)
    else
        set_anim_to_frame(m, 6)
    end
    
    m.actionTimer = m.actionTimer + 1
    if m.wall == nil and m.actionTimer > 2 then
        mario_set_forward_vel(m, 0.0)
        if (m.playerIndex == 0) then
            if (m.input & INPUT_Z_DOWN) ~= 0 then
                -- This caps the fall hspeed to the same hspeed as if the character jumped from a held dash
                s.jump_dash_state = true
                s.jumped_from_ground_dash = false
            end
        end
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    local stickAngle = m.intendedYaw
    local wallYaw = atan2s(m.wallNormal.z, m.wallNormal.x)
    local angleDiff = math.abs(stickAngle - wallYaw)
    if angleDiff > 32768 then
        angleDiff = 2 * 32768 - angleDiff
    end

    local normalStickMag = m.controller.stickMag / 64

    if (normalStickMag < 0.1 or angleDiff < 21500) then
        mario_set_forward_vel(m, 0.0)
        m.vel.y = 0
        if (m.playerIndex == 0) then
            if (m.input & INPUT_Z_DOWN) ~= 0 then
                -- This caps the fall hspeed to the same hspeed as if the character jumped from a held dash
                s.jump_dash_state = true
                s.jumped_from_ground_dash = false
            end
        end
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    return 0
end

function act_wall_slide_gravity(m)
    m.vel.y = m.vel.y - 3

    if m.vel.y < -20 then
        m.vel.y = -20
    end
end




--- DASH
-- For now, only X uses this action. But it should be a common action to other charcters added in a future update
function act_dash(m)
    if (m.actionTimer == 0) then
        mmx_play_sound(SOUND_X_DASH, m.pos, 0.5)
    end
    local s = gPlayerSyncTable[m.playerIndex]
    m.marioBodyState.headAngle.z = 0
    m.marioBodyState.headAngle.x = 0
    set_character_animation(m, CHAR_ANIM_SINGLE_JUMP)
    if s.x_shooting == 0 or s.vile_shooting == 0 then
        smlua_anim_util_set_animation(m.marioObj, "x_dash")
    end
    local mGfx = m.marioObj.header.gfx
    mario_set_forward_vel(m, 62.0)
    local performStep = perform_ground_step(m)
    if performStep == GROUND_STEP_LEFT_GROUND then
        set_mario_action(m, ACT_FREEFALL_CUSTOM, 0)
    elseif performStep == GROUND_STEP_NONE then
        align_with_floor(m)
        set_mario_particle_flags(m, PARTICLE_DUST, false)
        spawn_non_sync_object(id_bhvTrail, E_model_x_alpha_dash, mGfx.pos.x, mGfx.pos.y + 15, mGfx.pos.z, function (obj)
            obj.oMoveAngleYaw = m.faceAngle.y
        end)
    --elseif performStep == GROUND_STEP_HIT_WALL then
    --    set_mario_action(m, ACT_WALKING, 0)
    end
    if (m.input & INPUT_A_PRESSED) ~= 0 then
        if (m.playerIndex == 0) then
            s.jumped_from_ground_dash = true
            s.jump_dash_state = true
            s.ground_dash_cooldown = 5
        end
        return set_mario_action(m, ACT_JUMP, 0)
    end

    if m.actionTimer >= 13 or ((m.input & INPUT_Z_DOWN) == 0 and m.actionTimer >= 3) then
        if (m.playerIndex == 0) then
            if mario_floor_is_steep(m) ~= 0 then
                s.ground_dash_cooldown = 8
            else
                s.ground_dash_cooldown = 5
            end
        end
        mario_set_forward_vel(m, 30.0)
        return set_mario_action(m, ACT_WALKING, 0)
    end
    m.actionTimer = m.actionTimer + 1

    return 0
end




--- AIR DASH
-- For now, only X uses this action. But it should be a common action to other charcters added in a future update
function act_air_dash(m)
    if (m.actionTimer == 0) then
        mmx_play_sound(SOUND_X_DASH, m.pos, 0.5)
    end
    local s = gPlayerSyncTable[m.playerIndex]
    local mmx = gMmxStates[m.playerIndex]
    --set_character_animation(m, CHAR_ANIM_SINGLE_JUMP)
    s.already_airdashed = true
    if s.x_shooting == 0 or s.vile_shooting == 0 then
        smlua_anim_util_set_animation(m.marioObj, "x_dash")
    end
    m.marioObj.header.gfx.pos.y = m.pos.y
    local mGfx = m.marioObj.header.gfx
    
    spawn_non_sync_object(id_bhvTrail, E_model_x_alpha_dash, mGfx.pos.x , mGfx.pos.y + 45 , mGfx.pos.z , function (obj)
        obj.oMoveAngleYaw = m.faceAngle.y
    end)

    mario_set_forward_vel(m, 62.0)

    if m.actionTimer >= 13 or ((m.input & INPUT_Z_DOWN) == 0 and m.actionTimer >= 3) then
        mmx.smoothContinuityFlag = true
        mario_set_forward_vel(m, 32.0)
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    m.marioBodyState.headAngle.z = 0
    m.marioBodyState.headAngle.x = 0
    m.actionTimer = m.actionTimer + 1
    
    m.vel.y = 0

    common_air_action_step(m, ACT_FREEFALL_LAND_STOP, MARIO_ANIM_SLIDEJUMP,
                           AIR_STEP_CHECK_HANG)

    return false
end


hook_mario_action(ACT_DASH,                         { every_frame = act_dash })
hook_mario_action(ACT_AIR_DASH,                     { every_frame = act_air_dash })
hook_mario_action(ACT_AIR_HIT_WALL_CUSTOM,          { every_frame = act_air_hit_wall_custom })
hook_mario_action(ACT_WALL_SLIDE,                   { every_frame = act_wall_slide, gravity = act_wall_slide_gravity })
hook_mario_action(ACT_WALL_KICK_CUSTOM,             { every_frame = act_wall_kick_custom})
hook_mario_action(ACT_JUMP_CUSTOM,                  { every_frame = act_jump_custom})
hook_mario_action(ACT_FREEFALL_CUSTOM,              { every_frame = act_freefall_custom})
hook_mario_action(ACT_WALKING_CUSTOM,              { every_frame = act_walking_custom})








-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
---- UPDATE LOOP --------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

-- Functions

function executeSmoothContinuity(m)
    local mmx = gMmxStates[m.playerIndex]
    m.pos.x = m.pos.x + mmx.dPrevPos.x
    m.pos.y = m.pos.y + mmx.dPrevPos.y
    m.pos.z = m.pos.z + mmx.dPrevPos.z
    m.marioObj.header.gfx.pos.x = m.pos.x
    m.marioObj.header.gfx.pos.y = m.pos.y
    m.marioObj.header.gfx.pos.z = m.pos.z
end

-- Local player

function mmx_update_local_intro_timer(m)
    if (m.action == ACT_INTRO_CUTSCENE) and introTimer < 1195 then
        introTimer = introTimer + 1
    end
end

function mmx_local_update_air(m)
    local s = gPlayerSyncTable[m.playerIndex]
    -- (Automatically synced?)
    if ((m.action == ACT_JUMP_CUSTOM) 
    or (m.action == ACT_FREEFALL_CUSTOM) 
    or (m.action == ACT_WALL_KICK_CUSTOM) 
    or (m.action == ACT_TOP_OF_POLE_JUMP)
    or (m.action == ACT_WALL_KICK_AIR)) and s.just_walljumped == 0 then
        m.faceAngle.y = m.intendedYaw
    end

    if (m.action & ACT_FLAG_AIR) == 0 then
        s.firstWallKick = true
        s.quasiSameSideWallKicks = 0
    end
end

function mmx_local_update_landing(m)
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.action == ACT_JUMP_LAND or m.action == ACT_FREEFALL_LAND_STOP or m.action == ACT_FREEFALL_LAND) then
        s.land_timer = s.land_timer + 1
    else
        s.land_timer = 0
    end
end

function mmx_local_update_walljump(m)
    local s = gPlayerSyncTable[m.playerIndex]
    if s.just_walljumped > 0 then
        s.just_walljumped = s.just_walljumped - 1
    end
end


-- All players

function mmx_update_props(m)
    local mmx = gMmxStates[m.playerIndex]
    local mGfx = m.marioObj.header.gfx
    local s = gPlayerSyncTable[m.playerIndex]

    if (m.marioBodyState.capState == MARIO_HAS_DEFAULT_CAP_OFF) then
        m.marioBodyState.capState = MARIO_HAS_DEFAULT_CAP_ON
    end

    if mmx.smoothContinuityFlag then
        executeSmoothContinuity(m)
        mmx.smoothContinuityFlag = false
    end

    -- Offset the Y model because X/Vile is taller than Mario
    m.marioObj.hookRender = true
    if (m.playerIndex == 0) then
        if (m.action ~= ACT_GROUND_POUND
        and m.action ~= ACT_CROUCHING 
        and m.action ~= ACT_START_CRAWLING 
        and m.action ~= ACT_CRAWLING 
        and m.action ~= ACT_STOP_CRAWLING
        and m.action ~= ACT_QUICKSAND_DEATH) then
            mGfx.pos.y = m.pos.y + MODEL_Y_OFFSET
        end
        if (m.action == ACT_CROUCHING) then
            mGfx.pos.y = m.pos.y + MODEL_Y_OFFSET - 10
        end
    end
    
    -- No fall damage
    m.peakHeight = m.pos.y
end

function mmx_update_sleeping(m)
    if (m.action == ACT_SLEEPING or m.action == ACT_START_SLEEPING) then
        set_mario_action(m, ACT_IDLE, 0)
    end
end

function mmx_update_water_jump(m)
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.action == ACT_WATER_JUMP) then
        smlua_anim_util_set_animation(m.marioObj, s.character .. "_jump")
        if (m.vel.y <= 1) then
            return set_mario_action(m, ACT_FREEFALL_CUSTOM, 0)
        end
    end
end

function mmx_update_idle(m, shooting)
    local s = gPlayerSyncTable[m.playerIndex]
    if m.action == ACT_IDLE then
        -- MMX moveset characters don't fall asleep
        m.actionState = 0
        if shooting == 0 and s.character ~= 'none' then
            smlua_anim_util_set_animation(m.marioObj, s.character .. "_idle")
        end
    end

    if m.action == ACT_WARP_DOOR_SPAWN or m.action == ACT_READING_NPC_DIALOG or m.action == ACT_READING_SIGN and s.character ~= 'none' then
        smlua_anim_util_set_animation(m.marioObj, s.character .. "_idle")
    end
end

function mmx_update_walking(m, shooting)
    local mGfx = m.marioObj.header.gfx
    local s = gPlayerSyncTable[m.playerIndex]
    if m.action == ACT_WALKING_CUSTOM or m.action == ACT_WALKING then
        local char_factor = 0.95
        if (s.character == 'vile') then
            if (m.forwardVel > 32) then
                char_factor = m.forwardVel / 32
            end
            char_factor = char_factor * 0.95
        end
        mGfx.animInfo.animAccel = math.floor(mGfx.animInfo.animAccel / 1.8 / char_factor)
        if shooting == 0 then
            --m.marioBodyState.torsoAngle.x = m.marioBodyState.torsoAngle.x / 1.8
            --m.marioBodyState.torsoAngle.z = m.marioBodyState.torsoAngle.z / 1.8
            smlua_anim_util_set_animation(m.marioObj, "x_running")
        end
    end
end

function mmx_update_jump_fall(m, targetHspeed, targetHspeedDashJump, targetHspeedHoldDashJump, shooting)
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.action == ACT_JUMP_CUSTOM or m.action == ACT_FREEFALL_CUSTOM or m.action == ACT_STEEP_JUMP or m.action == ACT_WALL_KICK_AIR or m.action == ACT_TOP_OF_POLE_JUMP) then

        local targetMaxHspeed = targetHspeed
        if s.jump_dash_state then
            if (s.jumped_from_ground_dash) then
                targetMaxHspeed = targetHspeedDashJump
            else
                targetMaxHspeed = targetHspeedHoldDashJump
            end
        end

        if (m.forwardVel >= targetMaxHspeed) then
            targetMaxHspeed = m.forwardVel
        end

        local normalStickMag = m.controller.stickMag / 64
        mario_set_forward_vel(m, targetMaxHspeed * normalStickMag)

        if shooting == 0 and s.character ~= 'none' then
            smlua_anim_util_set_animation(m.marioObj, s.character .. "_jump")
        end

        if m.vel.y <= 0 then
            local velFrame = math.floor(m.vel.y * 0.18)
            if velFrame < -9 then
                velFrame = -9
            end
            set_anim_to_frame(m, 12 + velFrame)
        end
    end
end

function mmx_update_butt_slide(m)
    if (m.action & ACT_FLAG_BUTT_OR_STOMACH_SLIDE) ~= 0 then
        m.marioBodyState.torsoAngle.x = 0
        m.marioBodyState.torsoAngle.z = 0
    end
end

function mmx_update_punching(m, shooting)
    local s = gPlayerSyncTable[m.playerIndex]
    local mmx = gMmxStates[m.playerIndex]
    if (m.controller.buttonPressed & Y_BUTTON) ~= 0 and (m.input & INPUT_A_DOWN) == 0 and not (m.action == ACT_PUNCHING or m.action == ACT_MOVE_PUNCHING) then
        if (m.action == ACT_IDLE) then
            mmx.execute_punch = true
            set_mario_action(m, ACT_PUNCHING, 0)
            mmx.execute_punch = false
        elseif (m.action == ACT_WALKING_CUSTOM or m.action == ACT_DECELERATING 
        or m.action == ACT_BRAKING or m.action == ACT_BRAKING_STOP or m.action == ACT_WALKING) then
            mmx.execute_punch = true
            set_mario_action(m, ACT_MOVE_PUNCHING, 0)
            mmx.execute_punch = false
        end
    end
    if (m.action == ACT_PUNCHING or m.action == ACT_MOVE_PUNCHING) then
        shooting = 0
        mario_check_object_grab(m)
        if (m.actionArg == 3) then
            set_mario_action(m, ACT_IDLE, 0)
        end
    end
end

function mmx_update_groundpound(m)
    local mmx = gMmxStates[m.playerIndex]
    if (m.controller.buttonPressed & Y_BUTTON) ~= 0 and 
    (m.action == ACT_JUMP_CUSTOM or m.action == ACT_FREEFALL_CUSTOM or m.action == ACT_WALL_KICK_CUSTOM or m.action == ACT_FLYING) then
        if m.action == ACT_FLYING and m.playerIndex == 0 then
            set_camera_mode(m.area.camera, m.area.camera.defMode, 1)
        end
        mmx.execute_groundpound = true
        set_mario_action(m, ACT_GROUND_POUND, 0)
        mmx.execute_groundpound = false
    end
end

function mmx_update_turnaround(m, shooting)
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.action == ACT_TURNING_AROUND) then
        if shooting == 0 then
            smlua_anim_util_set_animation(m.marioObj, "x_turnaround")
            local turningAnimFrame = 6 - math.floor(m.forwardVel / 4.6)
            if turningAnimFrame < 1 then
                turningAnimFrame = 1
            end
            set_anim_to_frame(m, turningAnimFrame)
        end
    end
end

function mmx_update_landing(m, shooting)
    local mGfx = m.marioObj.header.gfx
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.action == ACT_FREEFALL_LAND_STOP) then
        if shooting == 0 and s.character ~= 'none' then
            smlua_anim_util_set_animation(m.marioObj, s.character .. "_land")
        end
        mGfx.animInfo.animFrame = s.land_timer
    end
end

function mmx_update_brake_stop(m)
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.action == ACT_BRAKING_STOP) and s.character ~= 'none' then
        smlua_anim_util_set_animation(m.marioObj, s.character .. "_brake_stop")
    end
end

function mmx_update_stop_crouch(m)
    local s = gPlayerSyncTable[m.playerIndex]
    if (m.action == ACT_STOP_CROUCHING) and s.character ~= 'none' then
        smlua_anim_util_set_animation(m.marioObj, s.character .. "_stop_crouch")
    end
end